// T-Code Serial Controller v2.1
// TempestVR 12-8-20
// Please copy, share, learn, innovate, give attribution.
// This plugin outputs a stream of data to the Serial Port that can be
// used to drive a generic device (T-code) or a Robot Sex machine (RSM)
// 15-7-19, 1.0 - First release
// 28-8-19, 1.1 - Roll axis bug fix
// 16-12-19, 1.2 - Double-digit COM port issue fix
// 28-1-20, 1.3 - Update to TCode v0.2 (new axis definitions & new default baud rate: 115200)
// 1-3-20, 2.0 - Update to 5-DoF and 3-digit precision. Pulse frequency now automatic.
// 19-8-20, 2.1 - Update to 6-DoF 

using System.Linq;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System;
using System.IO.Ports; 

namespace SerialCtrl {
    public class TCodeCtrl : MVRScript {
     
        // Name of plugin
        public static string pluginName = "Serial Controller";

        // Serial Variables
        protected JSONStorableStringChooser portChooser;
        private string[] portNames;
        private string portSelected = "None";
        protected JSONStorableStringChooser baudChooser; 
        private string baudSelected = "115200";
        protected JSONStorableStringChooser modeChooser; 
        private string pulsestring = "";
        protected UIDynamicButton startButton;
        protected UIDynamicButton stopButton;
        private string outputMode = "T-code";
        protected JSONStorableString serialReport;
        protected UIDynamicTextField serialOutWindow;

        protected JSONStorableFloat OutMaxX;
        protected JSONStorableFloat OutMinX;
        protected JSONStorableFloat OutMaxY;
        protected JSONStorableFloat AdjustY;
        protected JSONStorableFloat OutMaxZ;
        protected JSONStorableFloat AdjustZ;
		protected JSONStorableFloat OutMaxRX;
        protected JSONStorableFloat AdjustRX;
        protected JSONStorableFloat OutMaxRY;
        protected JSONStorableFloat AdjustRY;
        protected JSONStorableFloat OutMaxRZ;
        protected JSONStorableFloat AdjustRZ;
        protected JSONStorableBool InvertX;
        protected JSONStorableBool InvertY;
        protected JSONStorableBool InvertZ;
		protected JSONStorableBool InvertRX;
        protected JSONStorableBool InvertRY;
        protected JSONStorableBool InvertRZ;
        protected JSONStorableBool OverrideBoolX;
        protected JSONStorableBool OverrideBoolY;
        protected JSONStorableBool OverrideBoolZ;
        protected JSONStorableBool OverrideBoolRX;
        protected JSONStorableBool OverrideBoolRY;
        protected JSONStorableBool OverrideBoolRZ;
        protected JSONStorableFloat OverrideX;
        protected JSONStorableFloat OverrideY;
        protected JSONStorableFloat OverrideZ;
		protected JSONStorableFloat OverrideRX;
        protected JSONStorableFloat OverrideRY;
        protected JSONStorableFloat OverrideRZ;

        // Measurement Variables
        protected JSONStorableStringChooser refChooser;
        protected JSONStorableStringChooser targetChooser;
        protected JSONStorableString referenceString;
        private Atom refObject;              // Male person
        protected JSONStorableString targetString;
        private Atom targObject;             // Female person
        protected JSONStorableStringChooser targetPointChooser;
        protected JSONStorableString targetPointString;
        private Vector3 targPointOff;
		private Vector3 targPointDir;
		private Vector3 targPointIndex;
        protected JSONStorableString positionReport;
        protected UIDynamicTextField positionWindow;
        protected UIDynamicButton positionResetButton;
        protected UIDynamicButton positionApplyButton;

        protected JSONStorableFloat RangeMaxX;
        protected JSONStorableFloat RangeMinX;
        protected JSONStorableFloat RangeMaxY;
        protected JSONStorableFloat RangeMinY;
        protected JSONStorableFloat RangeMaxZ;
		protected JSONStorableFloat TwistMaxRX;
		protected JSONStorableFloat TwistMinRX;
        protected JSONStorableFloat RollMaxRY;
        protected JSONStorableFloat RollMaxRZ;
        protected JSONStorableFloat RollMinRZ;
        protected JSONStorableFloat InduceLag;

        private FreeControllerV3 refObjA;    // Penis base
        private FreeControllerV3 refObjB;    // Penis tip
        private FreeControllerV3 refObjH;    // Head
        private Vector3 refA;                // Penis base position
        private Vector3 refB;                // Penis tip position
        private Vector3 refH;                // Head position
        private Vector3 refLineX;            // Penis reference (X) line
        private Vector3 refLineY;            // Penis left-right (Y) line
        private Vector3 refLineZ;            // Penis fwd-back (Z) line
        private FreeControllerV3 targObjHip; // Target Hip
        private Vector3 targHipX;            // Target hip position
        private Quaternion targHipRot;       // Target hip rotation
        private Vector3 targC;               // Target point position
        private float xTarget;               // x-coordinate of target relative to datum
        private float xMax = -1000f;         // Max recorded x-displacement of target
        private float xMin = 1000f;          // Min record x-displacement of target 
        private float xCmd;                  // Command value for X - L0
        private float yTarget;               // y-coordinate of target relative to datum
        private float yMax = -1000f;         // Max recorded y-displacement of target
        private float yMin = 1000f;          // Min record y-displacement of target 
        private float yCmd;                  // Command value for Y - L1
        private float zTarget;               // z-component of offset vector
        private float zMax = 0.05f;          // Max recorded z-displacement of target
        private float zMin = -0.05f;         // Min recorded z-displacement of target
        private float zCmd;                  // Command value for Z - L2
        private float rxTarget;              // x-coordinate of target relative to datum
        private float rxMax = 0.05f;         // Max recorded x-roll of target
        private float rxMin = -0.05f;        // Min record x-roll of target 
        private float rxCmd;                 // Command value for RX - R0
		private float ryTarget;              // y-coordinate of target relative to datum
        private float ryMax = 0.05f;         // Max recorded y-roll of target
        private float ryMin = -0.05f;        // Min record y-roll of target 
        private float ryCmd;                 // Command value for RY - R1
        private float rzTarget;              // z-component of offset vector
        private float rzMax = 0.05f;         // Max recorded z-roll of target
        private float rzMin = -0.05f;        // Min recorded z-roll of target
        private float rzCmd;                 // Command value for RZ - R2


        // Serial Stuff
        SerialPort serial;


        // Function to initiate plugin
        public override void Init() {
            try {

                pluginLabelJSON.val = "Serial Controller";

                // Left Column of UI - Serial stuff

                // Select COM Port
                portNames = System.IO.Ports.SerialPort.GetPortNames();
                List<string> comPorts = new List<string>(portNames);
                portChooser = new JSONStorableStringChooser("COM Port Chooser", comPorts, "None", "Select COM port", PortChooserCallback);
                UIDynamicPopup portChooserPopup = CreatePopup(portChooser, false);
		portChooserPopup.labelWidth = 300f;

                // Select Baud Rate
                List<string> baudRates = new List<string>();
                baudRates.Add("9600");
                baudRates.Add("19200");
                baudRates.Add("38400");
                baudRates.Add("74880");
                baudRates.Add("115200");
                baudRates.Add("230400");
                baudRates.Add("250000");
                baudChooser = new JSONStorableStringChooser("Baud Rate Chooser", baudRates, "115200", "Select baud rate", BaudChooserCallback);
                UIDynamicPopup baudChooserPopup = CreatePopup(baudChooser, false);
		baudChooserPopup.labelWidth = 300f;

                // Select Output Mode
                List<string> outputMode = new List<string>();
                outputMode.Add("T-code");
                outputMode.Add("RSM");
                modeChooser = new JSONStorableStringChooser("Output Mode Chooser", outputMode, "T-code", "Select output mode", OutputChooserCallback);
                UIDynamicPopup modeChooserPopup = CreatePopup(modeChooser, false);
		modeChooserPopup.labelWidth = 300f;

                // Serial buttons
                startButton = CreateButton("Start Serial", false);
                if (startButton != null) {
                    startButton.button.onClick.AddListener(StartButtonCallback);
                }
                stopButton = CreateButton("Stop Serial", false);
                if (stopButton != null) {
                    stopButton.button.onClick.AddListener(StopButtonCallback);
                }
 
                // Window to show what's being sent over serial
                serialReport = new JSONStorableString("Serial Output", "");
                serialOutWindow = CreateTextField(serialReport, false);

                // Axis limits
                // X
                OutMaxX = new JSONStorableFloat("Max X Output", 0.90f, 0.5f, 1f, true, true);
                RegisterFloat(OutMaxX);
                CreateSlider(OutMaxX, false);
                OutMinX = new JSONStorableFloat("Min X Output", 0.10f, 0f, 0.5f, true, true);
                RegisterFloat(OutMinX);
                CreateSlider(OutMinX, false);
                // Y                
                OutMaxY = new JSONStorableFloat("Max Y Output (+/-)", 0.5f, 0f, 0.5f, true, true);
                RegisterFloat(OutMaxY);
                CreateSlider(OutMaxY, false);
                AdjustY = new JSONStorableFloat("Adjust Y", 0f, -0.25f, 0.25f, true, true);
                RegisterFloat(AdjustY);
                CreateSlider(AdjustY, false);
                // Z                
                OutMaxZ = new JSONStorableFloat("Max Z Output (+/-)", 0.5f, 0f, 0.5f, true, true);
                RegisterFloat(OutMaxZ);
                CreateSlider(OutMaxZ, false);
                AdjustZ = new JSONStorableFloat("Adjust Z", 0f, -0.25f, 0.25f, true, true);
                RegisterFloat(AdjustZ);
                CreateSlider(AdjustZ, false);
				// RX                
                OutMaxRX = new JSONStorableFloat("Max RX Output (+/-)", 0.5f, 0f, 0.5f, true, true);
                RegisterFloat(OutMaxRX);
                CreateSlider(OutMaxRX, false);
                AdjustRX = new JSONStorableFloat("Adjust RX", 0f, -0.25f, 0.25f, true, true);
                RegisterFloat(AdjustRX);
                CreateSlider(AdjustRX, false);
                // RY                
                OutMaxRY = new JSONStorableFloat("Max RY Output (+/-)", 0.5f, 0f, 0.5f, true, true);
                RegisterFloat(OutMaxRY);
                CreateSlider(OutMaxRY, false);
                AdjustRY = new JSONStorableFloat("Adjust RY", 0f, -0.25f, 0.25f, true, true);
                RegisterFloat(AdjustRY);
                CreateSlider(AdjustRY, false);
                // RZ                
                OutMaxRZ = new JSONStorableFloat("Max RZ Output (+/-)", 0.5f, 0f, 0.5f, true, true);
                RegisterFloat(OutMaxRZ);
                CreateSlider(OutMaxRZ, false);
                AdjustRZ = new JSONStorableFloat("Adjust RZ", 0f, -0.25f, 0.25f, true, true);
                RegisterFloat(AdjustRZ);
                CreateSlider(AdjustRZ, false);


                // Axis inverters
                InvertX = new JSONStorableBool("Invert X-Axis", false);
				CreateToggle(InvertX, false);
                InvertY = new JSONStorableBool("Invert Y-Axis", false);
				CreateToggle(InvertY, false);
                InvertZ = new JSONStorableBool("Invert Z-Axis", false);
				CreateToggle(InvertZ, false);
				InvertRX = new JSONStorableBool("Invert X-Rotation", false);
				CreateToggle(InvertRX, false);
                InvertRY = new JSONStorableBool("Invert Y-Rotation", false);
				CreateToggle(InvertRY, false);
                InvertRZ = new JSONStorableBool("Invert Z-Rotation", false);
				CreateToggle(InvertRZ, false);

                // Axis manual overrides
                // X
                OverrideBoolX = new JSONStorableBool("Override X-Axis", false);
				CreateToggle(OverrideBoolX, false);
                OverrideX = new JSONStorableFloat("Override X Position", 0.5f, 0f, 1f, true, true);
                RegisterFloat(OverrideX);
                CreateSlider(OverrideX, false);
                // Y
                OverrideBoolY = new JSONStorableBool("Override Y-Axis", false);
				CreateToggle(OverrideBoolY, false);
                OverrideY = new JSONStorableFloat("Override Y Position", 0.5f, 0f, 1f, true, true);
                RegisterFloat(OverrideY);
                CreateSlider(OverrideY, false);
                // Z
                OverrideBoolZ = new JSONStorableBool("Override Z-Axis", false);
				CreateToggle(OverrideBoolZ, false);
                OverrideZ = new JSONStorableFloat("Override Z Position", 0.5f, 0f, 1f, true, true);
                RegisterFloat(OverrideZ);
                CreateSlider(OverrideZ, false);  
				// RX
                OverrideBoolRX = new JSONStorableBool("Override X-Rotation", false);
				CreateToggle(OverrideBoolRX, false);
                OverrideRX = new JSONStorableFloat("Override X Rotation", 0.5f, 0f, 1f, true, true);
                RegisterFloat(OverrideRX);
                CreateSlider(OverrideRX, false);
                // RY
                OverrideBoolRY = new JSONStorableBool("Override Y-Rotation", false);
				CreateToggle(OverrideBoolRY, false);
                OverrideRY = new JSONStorableFloat("Override Y Rotation", 0.5f, 0f, 1f, true, true);
                RegisterFloat(OverrideRY);
                CreateSlider(OverrideRY, false);
                // RZ
                OverrideBoolRZ = new JSONStorableBool("Override Z-Rotation", false);
				CreateToggle(OverrideBoolRZ, false);
                OverrideRZ = new JSONStorableFloat("Override Z Rotation", 0.5f, 0f, 1f, true, true);
                RegisterFloat(OverrideRZ);
                CreateSlider(OverrideRZ, false);  


                // Right Column of UI - Target tracking stuff

                // Make a list of 
                Atom currentAtom;
                List<string> targetChoices = new List<string>();
                foreach (string atomUID in SuperController.singleton.GetAtomUIDs()) {
                    currentAtom = SuperController.singleton.GetAtomByUid(atomUID);
                    if (atomUID != null && currentAtom.type == "Person") {
                        targetChoices.Add(atomUID);
                    }
                }

                // Select reference atom
                referenceString = new JSONStorableString("Male Person","Person#2");
                RegisterString(referenceString);
                if (SuperController.singleton.GetAtomByUid(referenceString.val) == null) referenceString.val = targetChoices[0];
                refObject = SuperController.singleton.GetAtomByUid(referenceString.val);
                refObjA = refObject.GetStorableByID("penisBaseControl") as FreeControllerV3;
                refObjB = refObject.GetStorableByID("penisTipControl") as FreeControllerV3;
                refObjH = refObject.GetStorableByID("headControl") as FreeControllerV3;
                refChooser = new JSONStorableStringChooser("Reference Chooser", targetChoices, referenceString.val, "Select Male", RefChooserCallback);
                UIDynamicPopup refChooserPopup = CreatePopup(refChooser, true);
				refChooserPopup.labelWidth = 300f;

                // Select target atom
                targetString = new JSONStorableString("Female Person","Person");
                RegisterString(targetString);
                if (SuperController.singleton.GetAtomByUid(targetString.val) == null) targetString.val = targetChoices[0];
                targObject = SuperController.singleton.GetAtomByUid(targetString.val);
                targetPointString = new JSONStorableString("Target Point","hipControl");
                RegisterString(targetPointString);
				targPointOff = new Vector3(0f,-1f,0f);
                targPointDir = new Vector3(0f,-1f,0f);
				targPointDir = targPointDir/targPointDir.magnitude;
				targPointIndex = new Vector3(0f,0f,-1f);
				targPointIndex = targPointIndex/targPointIndex.magnitude;
                targObjHip = targObject.GetStorableByID(targetPointString.val) as FreeControllerV3;
                targetChooser = new JSONStorableStringChooser("Target Chooser", targetChoices, targetString.val, "Select Female", TargetChooserCallback);
                UIDynamicPopup targetChooserPopup = CreatePopup(targetChooser, true);
				targetChooserPopup.labelWidth = 300f;


                // Select what body part to track
                List<string> targetPoints = new List<string>();
                targetPoints.Add("Hips");
                targetPoints.Add("Pelvis");
                targetPoints.Add("Head");
                targetPoints.Add("Left Hand");
                targetPoints.Add("Right Hand");
                targetPointChooser = new JSONStorableStringChooser("Target Point Chooser", targetPoints, "Hips", "Select target point", TargetPointChooserCallback);
                UIDynamicPopup targetPointChooserPopup = CreatePopup(targetPointChooser, true);
				targetPointChooserPopup.labelWidth = 300f;

                
                // Window to show target position
                positionReport = new JSONStorableString("Target Position", "");
                positionWindow = CreateTextField(positionReport, true);

                // Position tracker buttons
                positionResetButton = CreateButton("Reset Recorded Limits", true);
                if (positionResetButton != null) {
                    positionResetButton.button.onClick.AddListener(PositionResetButtonCallback);
                }
                positionApplyButton = CreateButton("Apply Recorded Limits", true);
                if (positionApplyButton != null) {
                    positionApplyButton.button.onClick.AddListener(PositionApplyButtonCallback);
                }

                // Set defined measurement zone limits
                RangeMaxX = new JSONStorableFloat("X Range Max", 1f, -2f, 3f, true, true);
                RegisterFloat(RangeMaxX);
                CreateSlider(RangeMaxX, true);  
                RangeMinX = new JSONStorableFloat("X Range Min", 0f, -2f, 3f, true, true);
                RegisterFloat(RangeMinX);
                CreateSlider(RangeMinX, true);  
                RangeMaxY = new JSONStorableFloat("Y Range Max", 0.5f, -2f, 2f, true, true);
                RegisterFloat(RangeMaxY);
                CreateSlider(RangeMaxY, true);  
                RangeMinY = new JSONStorableFloat("Y Range Min", -0.5f, -2f, 2f, true, true);
                RegisterFloat(RangeMinY);
                CreateSlider(RangeMinY, true);  
                RangeMaxZ = new JSONStorableFloat("Z Range Max (+/-)", 0.4f, 0f, 0.6f, true, true);
                RegisterFloat(RangeMaxZ);
                CreateSlider(RangeMaxZ, true); 
				TwistMaxRX = new JSONStorableFloat("X Twist Max", 2f, -3.14159f, 3.14159f, true, true);
                RegisterFloat(TwistMaxRX);
                CreateSlider(TwistMaxRX, true); 
                TwistMinRX = new JSONStorableFloat("X Twist Min", -2f, -3.14159f, 3.14159f, true, true);
                RegisterFloat(TwistMinRX);
                CreateSlider(TwistMinRX, true); 
                RollMaxRY = new JSONStorableFloat("Y Roll Max (+/-)", 0.4f, 0f,1f, true, true);
                RegisterFloat(RollMaxRY);
                CreateSlider(RollMaxRY, true); 
                RollMaxRZ = new JSONStorableFloat("Z Roll Max", 0.8f, -1f, 1f, true, true);
                RegisterFloat(RollMaxRZ);
                CreateSlider(RollMaxRZ, true); 
                RollMinRZ = new JSONStorableFloat("Z Roll Min", -0.8f, -1f, 1f, true, true);
                RegisterFloat(RollMinRZ);
                CreateSlider(RollMinRZ, true); 

                // Slider to allow adding lag to robot response 
	        InduceLag = new JSONStorableFloat("Induce Lag", 0.5f, 0f, 1f, true, true);
                RegisterFloat(InduceLag);
                CreateSlider(InduceLag, true); 


            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }



        // Update is called with each rendered frame by Unity
        void Update() {
            try {
                // Get current frame information
                refA = refObjA.transform.position;
                refB = refObjB.transform.position;
                refH = refObjH.transform.position;
                targHipX = targObjHip.transform.position;
                targHipRot = targObjHip.transform.rotation;

                // Calculate reference coordinate system
                refLineX = refB-refA;
                float refMag;  // reference magnitude, ie penis length
                refMag = refLineX.magnitude;
                refLineZ = Vector3.Cross(refLineX,refH-refA);
                refLineZ = (refMag/refLineZ.magnitude)*refLineZ;
                refLineY = Vector3.Cross(refLineX,refLineZ);
                refLineY = (refMag/refLineY.magnitude)*refLineY;

                // Calculate reference point position 1x penis length below hipController
                Vector3 offsetC,dirC,indexC;
                offsetC = refMag*targPointOff;
                offsetC = targHipRot*offsetC;
                targC = targHipX + offsetC; //(targHipRot*offsetC);
				dirC = targHipRot*targPointDir;
				indexC = targHipRot*targPointIndex;

                // Calculate robot X-axis (Up-Down)
                xTarget = Vector3.Dot(refLineX,targC-refA);
                xTarget = xTarget/(refMag*refMag);
                if (xTarget > xMax) xMax = xTarget;
                if (xTarget < xMin) xMin = xTarget;
                positionReport.val = " Target Position\n Min, Cur, Max\n X " + xMin.ToString("0.00") + ", " + xTarget.ToString("0.00") + ", " + xMax.ToString("0.00");

                // Calculate robot Y-axis (fwd-aft)
                yTarget = Vector3.Dot(refLineY,targC-refA);
                yTarget = yTarget/(refMag*refMag);
                if (yTarget > yMax) yMax = yTarget;
                if (yTarget < yMin) yMin = yTarget;
                positionReport.val = positionReport.val + "\n Y " + yMin.ToString("0.00") + ", " + yTarget.ToString("0.00") + ", " + yMax.ToString("0.00");

                // Calculate robot Z-axis (left-right)
                zTarget = Vector3.Dot(refLineZ,targC-refA);
                zTarget = zTarget/(refMag*refMag);
                if (zTarget > zMax) zMax = zTarget;
                if (zTarget < zMin) zMin = zTarget;
                positionReport.val = positionReport.val + "\n Z " + zMin.ToString("0.00") + ", " + zTarget.ToString("0.00") + ", " + zMax.ToString("0.00");

                // Calculate robot RX-axis (left-right roll)
				Vector3 clockRef;
				clockRef = indexC - (Vector3.Dot(refLineX/refMag,indexC)/refMag)*refLineX;
				rxTarget = Vector3.Dot(refLineY/refMag,clockRef/clockRef.magnitude);
				rxTarget = Convert.ToSingle(Math.Acos(rxTarget));
				if (Vector3.Dot(refLineZ,indexC)>0) {rxTarget = -rxTarget;}
                if (rxTarget > rxMax) rxMax = rxTarget;
                if (rxTarget < rxMin) rxMin = rxTarget;
				positionReport.val = positionReport.val + "\n RX " + rxMin.ToString("0.00") + ", " + rxTarget.ToString("0.00") + ", " + rxMax.ToString("0.00");
           
				// Calculate robot RY-axis (left-right roll)
                ryTarget = Vector3.Dot(refLineZ,dirC);
                ryTarget = ryTarget/refMag;
                if (ryTarget > ryMax) ryMax = ryTarget;
                if (ryTarget < ryMin) ryMin = ryTarget;
                positionReport.val = positionReport.val + "\n RY " + ryMin.ToString("0.00") + ", " + ryTarget.ToString("0.00") + ", " + ryMax.ToString("0.00");
           
                // Calculate robot RZ-axis (fwd-backward roll)
                rzTarget = Vector3.Dot(refLineY,dirC);
                rzTarget = rzTarget/refMag;
                if (rzTarget > rzMax) rzMax = rzTarget;
                if (rzTarget < rzMin) rzMin = rzTarget;
                positionReport.val = positionReport.val + "\n RZ " + rzMin.ToString("0.00") + ", " + rzTarget.ToString("0.00") + ", " + rzMax.ToString("0.00");
           
                // Raw channel commands from target position
                float xCmdRaw, yCmdRaw, zCmdRaw, rxCmdRaw, ryCmdRaw, rzCmdRaw; 
                xCmdRaw = OutMinX.val + (OutMaxX.val - OutMinX.val)*((xTarget-RangeMinX.val)/(RangeMaxX.val-RangeMinX.val));
                yCmdRaw = 0.5f - OutMaxY.val + 2*OutMaxY.val*((yTarget-RangeMinY.val)/(RangeMaxY.val-RangeMinY.val)) + AdjustY.val;
                zCmdRaw = 0.5f + OutMaxZ.val*(zTarget/RangeMaxZ.val) + AdjustZ.val;
				rxCmdRaw = 0.5f + OutMaxRX.val - 2*OutMaxRX.val*((rxTarget-TwistMinRX.val)/(TwistMaxRX.val-TwistMinRX.val)) + AdjustRX.val; 
                ryCmdRaw = 0.5f + OutMaxRY.val*(ryTarget/RollMaxRY.val) + AdjustRY.val;
                rzCmdRaw = 0.5f + OutMaxRZ.val - 2*OutMaxRZ.val*((rzTarget-RollMinRZ.val)/(RollMaxRZ.val-RollMinRZ.val)) + AdjustRZ.val;                

                // Add lag, if selected
                xCmd = InduceLag.val*xCmd + (1-InduceLag.val)*xCmdRaw;
                yCmd = InduceLag.val*yCmd + (1-InduceLag.val)*yCmdRaw;
                zCmd = InduceLag.val*zCmd + (1-InduceLag.val)*zCmdRaw;
				rxCmd = InduceLag.val*rxCmd + (1-InduceLag.val)*rxCmdRaw;
                ryCmd = InduceLag.val*ryCmd + (1-InduceLag.val)*ryCmdRaw;
                rzCmd = InduceLag.val*rzCmd + (1-InduceLag.val)*rzCmdRaw;
				if (float.IsNaN(rxCmd)) rxCmd = rxCmdRaw; //Anti-NaN measures required on twist axis. F-knows why!

                // Axis inverters
                if (InvertX.val) xCmd = 1f - xCmd;
                if (InvertY.val) yCmd = 1f - yCmd;
                if (InvertZ.val) zCmd = 1f - zCmd;
				if (InvertRX.val) rxCmd = 1f - rxCmd;
                if (InvertRY.val) ryCmd = 1f - ryCmd;
                if (InvertRZ.val) rzCmd = 1f - rzCmd;

                // Manual overrides
                if (OverrideBoolX.val) xCmd = OverrideX.val;
                if (OverrideBoolY.val) yCmd = OverrideY.val;
                if (OverrideBoolZ.val) zCmd = OverrideZ.val;
				if (OverrideBoolRX.val) rxCmd = OverrideRX.val;
                if (OverrideBoolRY.val) ryCmd = OverrideRY.val;
                if (OverrideBoolRZ.val) rzCmd = OverrideRZ.val;
                

                // If serial active send pulse
                if (serial != null && serial.IsOpen) {
                    SendPulse();
                }

            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }


        protected void PortChooserCallback(string s) {
            portSelected = s;
        }

        protected void BaudChooserCallback(string s) {
            baudSelected = s;
        }

        protected void OutputChooserCallback(string s) {
            outputMode = s;
            // Set T-code defaults
            if (s=="T-code") {
                OutMaxX.val = 1f;
                OutMinX.val = 0.15f;
                OutMaxY.val = 0.5f;
                OutMaxZ.val = 0.5f;
            }
            // Set RSM defaults
            if (s=="RSM") {
                OutMaxX.val = 0.85f;
                OutMinX.val = 0.15f;
                OutMaxY.val = 0.5f;
                OutMaxZ.val = 0.4f;
            }
        }

        protected void RefChooserCallback(string s) {
            // Reset limit tracking
            PositionResetButtonCallback();
            // Set new reference object and nodes
            referenceString.val = s;
            refObject = SuperController.singleton.GetAtomByUid(s);
            refObjA = refObject.GetStorableByID("penisBaseControl") as FreeControllerV3;
            refObjB = refObject.GetStorableByID("penisTipControl") as FreeControllerV3;
            refObjH = refObject.GetStorableByID("headControl") as FreeControllerV3;
        }
 
        protected void TargetChooserCallback(string s) {
            // Reset limit tracking
            PositionResetButtonCallback();
            // Set new target object and nodes
            targetString.val = s;
            targObject = SuperController.singleton.GetAtomByUid(s);
            targObjHip = targObject.GetStorableByID(targetPointString.val) as FreeControllerV3;
        }

        protected void TargetPointChooserCallback(string s) {
            // Reset limit tracking
            PositionResetButtonCallback();
            // Assign a new position point to track
            if (s == "Head") {
                targetPointString.val = "headControl";
				targPointOff = new Vector3(0f,-0.2f,0.6f);
                targPointDir = new Vector3(0f,-0.3f,0.9f);
				targPointIndex = new Vector3(0f,-0.9f,-0.3f);
            } else if (s == "Left Hand") {
                targetPointString.val = "lHandControl";
                targPointOff = new Vector3(0.5f,0f,0f);
				targPointDir = new Vector3(0f,0f,-1f);
				targPointIndex = new Vector3(-0.707f,-0.707f,0f);
            } else if (s == "Right Hand") {
                targetPointString.val = "rHandControl";
                targPointOff = new Vector3(0.5f,0f,0f);
				targPointDir = new Vector3(0f,0f,-1f);
				targPointIndex = new Vector3(-0.707f,-0.707f,0f);
            } else if (s == "Pelvis") {
                targetPointString.val = "pelvisControl";
                targPointOff = new Vector3(0f,-1f,0f);
				targPointDir = new Vector3(0f,-1f,0f);
				targPointIndex = new Vector3(0f,0f,-1f);
            } else {
                targetPointString.val = "hipControl";
                targPointOff = new Vector3(0f,-1f,0f);
				targPointDir = new Vector3(0f,-1f,0f);
				targPointIndex = new Vector3(0f,0f,-1f);
            }
			targPointDir = targPointDir/targPointDir.magnitude;
			targPointIndex = targPointIndex/targPointIndex.magnitude;
            targObjHip = targObject.GetStorableByID(targetPointString.val) as FreeControllerV3;
        }



        // Serial button functions
        protected void StartButtonCallback() {
            StartSerial();
            SuperController.LogMessage("Serial connection started: " + portSelected + ", baud " + baudSelected );
        }
        protected void StopButtonCallback() {
            StopSerial();
            SuperController.LogMessage("Serial connection stopped");
        }
        protected void PositionResetButtonCallback() {
            xMax = xTarget;
            xMin = xTarget;
            yMax = yTarget;
            yMin = yTarget;
            zMax = zTarget;
            zMin = zTarget;
            ryMax = ryTarget;
            ryMin = ryTarget;
            rzMax = rzTarget;
            rzMin = rzTarget;
        }
        protected void PositionApplyButtonCallback() {
            RangeMaxX.val = xMax;
            RangeMinX.val = xMin;
            RangeMaxY.val = yMax;
            RangeMinY.val = yMin;
            RangeMaxZ.val = zMax;
            if (zMax<-zMin) RangeMaxZ.val = -zMin;
			TwistMaxRX.val = rxMax;
            TwistMinRX.val = rxMin;
            RollMaxRY.val = ryMax;
            if (ryMax<-ryMin) RollMaxRY.val = -ryMin;
            RollMaxRZ.val = rzMax;
            RollMinRZ.val = rzMin;
            
        }


        // Function to start serial
        void StartSerial() {
            if (portSelected != "None") {
                // Add extra characters for COM ports > 9
                if (portSelected.Substring(0,3) == "COM") {
                    if (portSelected.Length != 4) {
                        portSelected = "\\\\.\\" + portSelected;
                    }
                }

                // Open the serial connection
                serial = new SerialPort(portSelected, Convert.ToInt32(baudSelected));
                serial.Open();
                serial.ReadTimeout = 10;
            }
        }
        // Function to stop serial
        void StopSerial() {
            if(serial != null && serial.IsOpen) serial.Close();
        }
        // Function to send serial
        void SendPulse() {
            if (outputMode == "T-code") {
                // Create T-code lines to send 3 linear axis positions "L0", "L1" & "L2"
                pulsestring = "L0" + TcodeNumber(xCmd,OutMinX.val,OutMaxX.val) + "\nL1" + TcodeNumber(yCmd,0.5f-OutMaxY.val,0.5f+OutMaxY.val) + "\nL2" + TcodeNumber(zCmd,0.5f-OutMaxZ.val,0.5f+OutMaxZ.val);
                // Add Tcode lines to send 2 rotation positions "R1" & "R2"
                pulsestring = pulsestring + "\nR0" + TcodeNumber(rxCmd,0.5f-OutMaxRX.val,0.5f+OutMaxRX.val) + "\nR1" + TcodeNumber(ryCmd,0.5f-OutMaxRY.val,0.5f+OutMaxRY.val) + "\nR2" + TcodeNumber(rzCmd,0.5f-OutMaxRZ.val,0.5f+OutMaxRZ.val);
                serialReport.val = " T-code:\n" + pulsestring;
                serial.WriteLine(pulsestring);
            } else if (outputMode == "RSM") {
                // RSM controlled by directly commanding servos through Pololu Maestro Servo Controller, 2 axes only
                float servo0f,servo1f;
                // Mix channels to send servo commands in the range 4000-8000 (quarter microseconds)
                // Note one servo inverted
                servo0f = 8000 - 4000*xCmd - 2000*(ryCmd-0.5f);   // Left servo
                servo1f = 4000 + 4000*xCmd - 2000*(ryCmd-0.5f);   // Right servo
                // Make sure servos stay within command range limit
                if (servo0f > 8000) servo0f = 8000;
                if (servo0f < 4000) servo0f = 4000;
                if (servo1f > 8000) servo1f = 8000;
                if (servo1f < 4000) servo1f = 4000;
                // Convert floats into unsigned integers
                UInt32 servo0,servo1;
                servo0 = Convert.ToUInt32(servo0f);
                servo1 = Convert.ToUInt32(servo1f);
                // Create 8-byte array to contain serial command
                byte[] serialBytes = new byte[8];
                serialBytes[0] = 0x84;                                    // Move command identifier
                serialBytes[1] = 0x00;                                    // Servo number (left servo - 0)
                serialBytes[2] = Convert.ToByte(servo0 & 0x7F);           // First 7 bits of 14-bit position command
                serialBytes[3] = Convert.ToByte((servo0 >> 7) & 0x7F);    // Second 7 bits of 14-bit position command
                serialBytes[4] = 0x84;                                    // Move command identifier
                serialBytes[5] = 0x01;                                    // Servo number (right servo - 1)
                serialBytes[6] = Convert.ToByte(servo1 & 0x7F);           // First 7 bits of 14-bit position command
                serialBytes[7] = Convert.ToByte((servo1 >> 7) & 0x7F);    // Second 7 bits of 14-bit position command
                serial.Write(serialBytes,0,serialBytes.Length);
                serialReport.val = " RSM:\n " + serialBytes[0].ToString("0") + " " + serialBytes[1].ToString("0") + " " + serialBytes[2].ToString("0")  + " " + serialBytes[3].ToString("0") + "\n " + serialBytes[4].ToString("0") + " " + serialBytes[5].ToString("0") + " " + serialBytes[6].ToString("0")  + " " + serialBytes[7].ToString("0"); 
            }
        }


        // Function to create t-code magnitude number
        string TcodeNumber(float input, float min, float max) {
            if (input > max) input = max;
            if (input < min) input = min;
            input = input*1000;
            string output;
            if (input >= 999f) {
                output = "999";
            } else if (input >= 1f) {
                output = input.ToString("000");
            } else {
                output = "000";
            }
            return output;
        }


        // Function to close plugin
        void OnDestroy() {
            try {
                StopSerial();
            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }


    }
}