// T-Code Random Stroker v0.4
// TempestVR 15-5-20
// Please copy, share, learn, innovate, give attribution.
// This plugin animates the hips of a female with a series of fun randomised motions.
// It also then outputs a synced stream of data to the Serial Port that can be used 
// to drive a generic device (T-code)
//
// 15-3-20, v0.4 - Rough WIP. Random stroker, T-Code output, animated pelvis. Limited control interface.

// Import libraries
using System.Linq;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System;
using System.IO.Ports; 

// The magic starts here
namespace SerialCtrl {
    public class RandomStroker : MVRScript {
		
        // Name of plugin
        public static string pluginName = "Random Stroker";
		
		// Serial Variables
        protected JSONStorableStringChooser portChooser;
        private string[] portNames;
        private string portSelected = "None";
		private int baudRate;
		private string pulseString = "";
        protected UIDynamicButton startButton;
        protected UIDynamicButton stopButton;
        protected JSONStorableString serialReport;
        protected UIDynamicTextField serialOutWindow;
		
		// TCode Variables
		protected JSONStorableFloat OutMaxX;
        protected JSONStorableFloat OutMinX;
        protected JSONStorableFloat OutMaxRY;
        protected JSONStorableFloat AdjustRY;
        protected JSONStorableFloat OutMaxRZ;
        protected JSONStorableFloat AdjustRZ;
		protected JSONStorableFloat OutMaxV0;
        protected JSONStorableBool InvertX;
        protected JSONStorableBool InvertRY;
        protected JSONStorableBool InvertRZ;
        protected JSONStorableBool OverrideBoolX;
        protected JSONStorableBool OverrideBoolRY;
        protected JSONStorableBool OverrideBoolRZ;
		protected JSONStorableBool OverrideBoolV0;
        protected JSONStorableFloat OverrideX;
        protected JSONStorableFloat OverrideRY;
        protected JSONStorableFloat OverrideRZ;
		protected JSONStorableFloat OverrideV0;
		
		// Output command variables
		private float xCmd;
		private float ryCmd;
		private float rzCmd;
		private float v0Cmd;
		
		// Stroker UI variables
		protected JSONStorableFloat levelControl;
		protected JSONStorableString strokerReport;
        protected UIDynamicTextField strokerWindow;
		protected JSONStorableBool vibeModes;
		
		// Random stroker variables
		// Timer
		private float step;             // Time step [sec]
		private float iMin = 0.5f;      // Interval Minimum [sec]
		private float iMax = 6f;        // Interval Maximum [sec]

		// Starting positions
		private float level = 0f;       // Start at minimum intensity
		private float timer = 0f;       // Interval countdown timer [sec] (time to reach new key point)
		private float xUp = 1f;         // Current stroke top [0.0 - 1.0]
		private float xDown = 0.8f;     // Current stroke bottom [0.0 - 1.0]
		private float xAngle = 0f;      // Current angular position of stroke [radians]
		private float grind = 0f;       // Current grind factor
		private float tilt = 0.0f;      // Current off axis tilt of MAxSR [0.0 - 1.0]
		private float tiltAngle = 0f;   // Current direction of tilt of MAxSR [radians]
		private float vibe1 = 0f;       // Current vibration level - channel 1 [0.0 - 1.0]
		private int mode = 3;           // Current movement mode [1,2,3,etc]
		
		// Rates
		private float xW = 0f;          // Current angular speed of stoke [radians/sec]
		private float tiltW = 0f;       // Current angular speed of tilt direction [radians/sec]

		// Change rates
		private float levelV = 0f;      // Rate of increase of intensity
		private float xDUp = 0f;        // Rate of change of current stroke top
		private float xDDown = 0f;      // Rate of change of current stroke bottom
		private float xDW = 0f;         // Rate of change of angular stroke speed [radians/sec^2]
		private float grindV = 0f;      // Rate of change of grind
		private float tiltV = 0f;       // Rate of change of tilt
		private float tiltDW = 0f;      // Rate of change of speed of tilt direction [radians/sec^2]
		private float vibe1V = 0f;      // Rate of change of vibration level

		
		// Atom Manipulation Variables
		// Male variables
		protected JSONStorableStringChooser maleChooser;
		protected JSONStorableString maleString;
        private Atom maleObject;
		private FreeControllerV3 malePenisTip;
        private FreeControllerV3 malePenisBase;
        private FreeControllerV3 maleHead;
        private Vector3 malePenisTipX;
        private Vector3 malePenisBaseX;
		private Quaternion malePenisRef;
		private float malePenisLength;
        private Vector3 malePenisRefX;
        private Vector3 malePenisRefY;
        private Vector3 malePenisRefZ;
        private Vector3 maleHeadX;

		// Female variables
		protected JSONStorableStringChooser femaleChooser;
        protected JSONStorableString femaleString;
        private Atom femaleObject; 
        private FreeControllerV3 femaleHead;
        private FreeControllerV3 femalePelvis;
		private FreeControllerV3 femaleThighLeft;
		private FreeControllerV3 femaleThighRight;
	
	
		
		// Serial Stuff
        SerialPort serial;
		
		// Function to initiate plugin
        public override void Init() {
			
			pluginLabelJSON.val = "Random Stroker";
			
			// Left Column of UI - Serial stuff
			
			// Select COM Port
            portNames = System.IO.Ports.SerialPort.GetPortNames();
            List<string> comPorts = new List<string>(portNames);
			portChooser = new JSONStorableStringChooser("COM Port Chooser", comPorts, "None", "Select COM port", PortChooserCallback);
			UIDynamicPopup portChooserPopup = CreatePopup(portChooser, false);
			portChooserPopup.labelWidth = 300f;
			
			// Set Baud Rate
			baudRate = 115200;
			
			// Serial buttons
			startButton = CreateButton("Start Serial", false);
			if (startButton != null) {
				startButton.button.onClick.AddListener(StartButtonCallback);
			}
			stopButton = CreateButton("Stop Serial", false);
			if (stopButton != null) {
				stopButton.button.onClick.AddListener(StopButtonCallback);
			}
			
			// Window to show what's being sent over serial
			serialReport = new JSONStorableString("Serial Output", "");
			serialOutWindow = CreateTextField(serialReport, false);
			
			// Axis limits
			// X
			OutMaxX = new JSONStorableFloat("Max X Output", 0.90f, 0.5f, 1f, true, true);
			RegisterFloat(OutMaxX);
			CreateSlider(OutMaxX, false);
			OutMinX = new JSONStorableFloat("Min X Output", 0.10f, 0f, 0.5f, true, true);
			RegisterFloat(OutMinX);
			CreateSlider(OutMinX, false);
			// RY
			OutMaxRY = new JSONStorableFloat("Max RY Output (+/-)", 0.4f, 0f, 0.5f, true, true);
			CreateSlider(OutMaxRY, false);
			AdjustRY = new JSONStorableFloat("Adjust RY", 0f, -0.25f, 0.25f, true, true);
			RegisterFloat(AdjustRY);
			CreateSlider(AdjustRY, false);
			// RZ
			OutMaxRZ = new JSONStorableFloat("Max RZ Output (+/-)", 0.4f, 0f, 0.5f, true, true);
			RegisterFloat(OutMaxRZ);
			CreateSlider(OutMaxRZ, false);
			AdjustRZ = new JSONStorableFloat("Adjust RZ", 0f, -0.25f, 0.25f, true, true);
			RegisterFloat(AdjustRZ);
			CreateSlider(AdjustRZ, false);
			// V0
			OutMaxV0 = new JSONStorableFloat("Max V0 Output", 0.5f, 0f, 1f, true, true);
			RegisterFloat(OutMaxV0);
			CreateSlider(OutMaxV0, false);
			
			// Axis inverters
			InvertX = new JSONStorableBool("Invert X-Axis", false);
			CreateToggle(InvertX, false);
			InvertRY = new JSONStorableBool("Invert Y-Rotation", false);
			CreateToggle(InvertRY, false);
			InvertRZ = new JSONStorableBool("Invert Z-Rotation", false);
			CreateToggle(InvertRZ, false);
			
			// Axis manual overrides
			// X
			OverrideBoolX = new JSONStorableBool("Override X-Axis", false);
			CreateToggle(OverrideBoolX, false);
			OverrideX = new JSONStorableFloat("Override X Position", 0.5f, 0f, 1f, true, true);
			RegisterFloat(OverrideX);
			CreateSlider(OverrideX, false);
			// RY
			OverrideBoolRY = new JSONStorableBool("Override Y-Rotation", false);
			CreateToggle(OverrideBoolRY, false);
			OverrideRY = new JSONStorableFloat("Override Y Rotation", 0.5f, 0f, 1f, true, true);
			RegisterFloat(OverrideRY);
			CreateSlider(OverrideRY, false);
			// RZ
			OverrideBoolRZ = new JSONStorableBool("Override Z-Rotation", false);
			CreateToggle(OverrideBoolRZ, false);
			OverrideRZ = new JSONStorableFloat("Override Z Rotation", 0.5f, 0f, 1f, true, true);
			CreateSlider(OverrideRZ, false);  
			// V0
			OverrideBoolV0 = new JSONStorableBool("Override Vibration", false);
			CreateToggle(OverrideBoolV0, false);
			OverrideV0 = new JSONStorableFloat("Override Vibration", 0.5f, 0f, 1f, true, true);
			CreateSlider(OverrideV0, false);  
			
			
			// Right Column of UI 
			
			// Make a list of
			Atom currentAtom;
			List<string> targetChoices = new List<string>();
			foreach (string atomUID in SuperController.singleton.GetAtomUIDs()) {
				currentAtom = SuperController.singleton.GetAtomByUid(atomUID);
				if (atomUID != null && currentAtom.type == "Person") {
					targetChoices.Add(atomUID);
				}
			}
			
			// Set male character nodes
			// Use default male character, or else first on list
			maleString = new JSONStorableString("Male Person","Person#2");
            RegisterString(maleString);
            if (SuperController.singleton.GetAtomByUid(maleString.val) == null) maleString.val = targetChoices[0];
			// Set nodes
			maleObject = SuperController.singleton.GetAtomByUid(maleString.val);
            malePenisTip = maleObject.GetStorableByID("penisTipControl") as FreeControllerV3;
			malePenisTip.currentPositionState = FreeControllerV3.PositionState.On;
			malePenisTip.currentRotationState = FreeControllerV3.RotationState.On;
            malePenisBase = maleObject.GetStorableByID("penisBaseControl") as FreeControllerV3;
			malePenisBase.currentPositionState = FreeControllerV3.PositionState.On;
			malePenisBase.currentRotationState = FreeControllerV3.RotationState.On;
			maleHead = maleObject.GetStorableByID("headControl") as FreeControllerV3;
			// Chooser menu
			maleChooser = new JSONStorableStringChooser("Reference Chooser", targetChoices, maleString.val, "Select Male",MaleChooserCallback);
            UIDynamicPopup maleChooserPopup = CreatePopup(maleChooser, true);
			maleChooserPopup.labelWidth = 300f;

			// Set female character nodes
			// Use default female character, or else first on list
            femaleString = new JSONStorableString("Female Person","Person");
            RegisterString(femaleString);
            if (SuperController.singleton.GetAtomByUid(femaleString.val) == null) femaleString.val = targetChoices[0];
			// Set nodes
            femaleObject = SuperController.singleton.GetAtomByUid(femaleString.val);
            femalePelvis = femaleObject.GetStorableByID("pelvisControl") as FreeControllerV3;
			femalePelvis.currentPositionState = FreeControllerV3.PositionState.On;
			femalePelvis.currentRotationState = FreeControllerV3.RotationState.On;
			femaleThighLeft = femaleObject.GetStorableByID("lThighControl") as FreeControllerV3;
			femaleThighLeft.currentPositionState = FreeControllerV3.PositionState.On;
			femaleThighRight = femaleObject.GetStorableByID("rThighControl") as FreeControllerV3;
			femaleThighRight.currentPositionState = FreeControllerV3.PositionState.On;
            femaleHead = femaleObject.GetStorableByID("headControl") as FreeControllerV3;
			femaleHead.currentPositionState = FreeControllerV3.PositionState.On;
			femaleHead.currentRotationState = FreeControllerV3.RotationState.On;
			// Chooser menu
			femaleChooser = new JSONStorableStringChooser("Target Chooser", targetChoices, femaleString.val, "Select Female", FemaleChooserCallback);
            UIDynamicPopup femaleChooserPopup = CreatePopup(femaleChooser, true);
			femaleChooserPopup.labelWidth = 300f;

			
		    // Window to show target position
            strokerReport = new JSONStorableString("Target Position", "");
            strokerWindow = CreateTextField(strokerReport, true);
			
			levelControl = new JSONStorableFloat("Intensity Level", 0.5f, 0f, 1f, true, true);
			RegisterFloat(levelControl);
			CreateSlider(levelControl, true);
			
			vibeModes = new JSONStorableBool("Include Vibration Modes", false);
			CreateToggle(vibeModes, true);
			
		
        }
		
		
		
		// Update is called with each rendered frame by Unity
        void Update() {
			
			// Do the random stroker stuff
			
			// If interval countdown timer expires, reset
			if (timer <= 0) {
				
				// Mode change randomiser
				if (UnityEngine.Random.Range(0f, 1f) < 0.3) {
					int modeMax;
					if (vibeModes.val == true) {
						modeMax = 7;
					} else {
						modeMax = 5;
					}
					
					mode = UnityEngine.Random.Range(0,modeMax);
				}
				
				// Set new interval countdown timer
				timer = UnityEngine.Random.Range(iMin, iMax);
				
				// Stroke randomiser variables
				SetStrokeParams(mode);
				
			}

			// Set timestep
			step = Time.deltaTime;
			timer -= step;

			// Update operating values for this timestep
			level += levelV * step;
			xUp += xDUp * step;
			xDown += xDDown * step;
			xW += xDW * step;
			grind += grindV * step;
			tiltW += tiltDW * step;
			tilt += tiltV * step;
			vibe1 += vibe1V * step;
			xAngle += xW * step;
			tiltAngle += tiltW * step;

			// Calculate channel output values
			xCmd = (xUp + xDown)/2 + Mathf.Cos(xAngle)*(xUp - xDown)/2;
			ryCmd = 0.5f * tilt * Mathf.Sin(tiltAngle) + 0.5f;
			rzCmd = 0.5f * tilt * Mathf.Cos(tiltAngle) + 0.5f * grind * Mathf.Sin(xAngle)+ 0.5f;
			v0Cmd = vibe1;
			
			// Axis inverters
			if (InvertX.val) xCmd = 1f - xCmd;
			if (InvertRY.val) ryCmd = 1f - ryCmd;
			if (InvertRZ.val) rzCmd = 1f - rzCmd;
			
			// Manual overrides
			if (OverrideBoolX.val) xCmd = OverrideX.val;
			if (OverrideBoolRY.val) ryCmd = OverrideRY.val;
			if (OverrideBoolRZ.val) rzCmd = OverrideRZ.val;
			if (OverrideBoolV0.val) v0Cmd = OverrideV0.val;
			
			// Send data over serial
			if (serial != null && serial.IsOpen) {
				SendPulse();
			}
			
			// Do the animation stuff
            // Penis reference frame  -  UGLY. Need to fix this!
            malePenisRefZ = malePenisTip.transform.position - malePenisBase.transform.position;
            malePenisRefX = Vector3.Cross(malePenisRefZ,maleHead.transform.position-malePenisBase.transform.position);
            malePenisRefX = (malePenisRefZ.magnitude/malePenisRefX.magnitude)*malePenisRefX;
            malePenisRefY = Vector3.Cross(malePenisRefX,malePenisRefZ);
            malePenisRefY = (malePenisRefZ.magnitude/malePenisRefY.magnitude)*malePenisRefY;
			
			// New reference frame - Bette!
			malePenisRef = Quaternion.LookRotation((malePenisTip.transform.position - malePenisBase.transform.position),(maleHead.transform.position - malePenisBase.transform.position));
			malePenisLength = (malePenisTip.transform.position - malePenisBase.transform.position).magnitude;

			// Need to fix this too!
            femalePelvis.transform.rotation = Quaternion.LookRotation(malePenisRefY+(1.5f*(rzCmd-0.5f)*malePenisRefZ),malePenisRefZ-(1.5f*(rzCmd-0.5f)*malePenisRefY)-(1f*(ryCmd-0.5f)*malePenisRefX));
     
			// Animate the pelvis and thighs
			femalePelvis.transform.position = malePenisTip.transform.position + malePenisRef* new Vector3(0f,0.1f*(1f-xCmd),-0.075f + 0.18f*xCmd) + femalePelvis.transform.rotation* new Vector3(0f,0.05f,-0.02f);
			femaleThighLeft.transform.position = femalePelvis.transform.position + femalePelvis.transform.rotation* new Vector3(-0.08f,-0.05f - 0.05f*xCmd,0f);
			femaleThighRight.transform.position = femalePelvis.transform.position + femalePelvis.transform.rotation* new Vector3(0.08f,-0.05f - 0.05f*xCmd,0f);
			
			// Don't need this yet!
            //femaleHead.transform.position = malePenisBase.transform.position + 2f*malePenisRefZ + new Vector3(0f,0.6f,0f);
			
		}
		
		
		
		// Button callbacks
		protected void PortChooserCallback(string s) {
            portSelected = s;
        }
		protected void StartButtonCallback() {
			StartSerial();
			SuperController.LogMessage("Serial connection started: " + portSelected + ", baud " + baudRate.ToString() );
        }
		protected void StopButtonCallback() {
			StopSerial();
			SuperController.LogMessage("Serial connection stopped");
		}
		protected void MaleChooserCallback(string s) {
			// Set new male object and nodes
			maleString.val = s;
			maleObject = SuperController.singleton.GetAtomByUid(maleString.val);
            malePenisTip = maleObject.GetStorableByID("penisTipControl") as FreeControllerV3;
			malePenisTip.currentPositionState = FreeControllerV3.PositionState.On;
			malePenisTip.currentRotationState = FreeControllerV3.RotationState.On;
            malePenisBase = maleObject.GetStorableByID("penisBaseControl") as FreeControllerV3;
			malePenisBase.currentPositionState = FreeControllerV3.PositionState.On;
			malePenisBase.currentRotationState = FreeControllerV3.RotationState.On;
			maleHead = maleObject.GetStorableByID("headControl") as FreeControllerV3;
        }
 
        protected void FemaleChooserCallback(string s) {
			// Set new female object and nodes
			femaleString.val = s;
			
        }
		
		
		// Serial functions
		// Function to start serial
		void StartSerial() {
			if (portSelected != "None") {
				// Add extra characters for COM ports > 9
				if (portSelected.Substring(0,3) == "COM") {
					if (portSelected.Length != 4) {
						portSelected = "\\\\.\\" + portSelected;
					}
				}
				
				// Open the serial connection
				serial = new SerialPort(portSelected, baudRate);
				serial.Open();
				serial.ReadTimeout = 10;
			}
		}
		// Function to stop serial
        void StopSerial() {
			if(serial != null && serial.IsOpen) serial.Close();
		}
		// Function to send serial
        void SendPulse() {
			pulseString = TCode("L0",xCmd,OutMinX.val,OutMaxX.val);
			pulseString += " ";
			pulseString += TCode("R1",ryCmd,0.5f-OutMaxRY.val,0.5f+OutMaxRY.val);
			pulseString += " ";
			pulseString += TCode("R2",rzCmd,0.5f-OutMaxRZ.val,0.5f+OutMaxRZ.val);
			pulseString += " ";
			pulseString += TCode("V0",v0Cmd,0f,OutMaxV0.val);
			pulseString += "\n";
			serialReport.val = " T-code:\n" + pulseString;
            serial.WriteLine(pulseString);
		}
		
		
		// TCode function
        string TCode(string channel, float input, float min, float max) {
            input = min + input*(max-min);
            input = input*1000;
            string output;
            if (input >= 999f) {
                output = channel + "999";
            } else if (input >= 1f) {
                output = channel + input.ToString("000");
            } else {
                output = channel + "000";
            }
            return output;
        }
		
		// This function sets the new stroke parameters
		void SetStrokeParams(int setMode) {
			level = levelControl.val;
				
			// Temporary variables
			float stroke;
			float xDownNew;
			float xUpNew;
			float xWNew;
			float grindNew;
			float tiltNew;
			float tiltWNew;
			float vibe1New;
			
			// Movement modes are defined here
			if (setMode == 6) {
				// Vibe Grind
				strokerReport.val = "Vibe Grind";
				stroke = ScaledRandom(0.1f, 0.2f, 0.2f, 0.3f, level);
				xDownNew = UnityEngine.Random.Range(0f, 0.1f);
				xUpNew = xDownNew + stroke;
				xWNew = ScaledRandom(1f, 2f, 2f, 4f, level);
				grindNew = ScaledRandom(-0.9f, 0.9f, -0.9f, 0.9f, level);
				tiltNew = ScaledRandom(0f, 0.1f, 0f, 0.1f, level);
				tiltWNew = ScaledRandom(-1f, 1f, -2f, 2f, level);
				vibe1New = ScaledRandom(0.1f, 0.2f, 0.3f, 1f, level);
			} else if (setMode == 5) {
				// Vibe Tease
				strokerReport.val = "Vibe Tease";
				stroke = ScaledRandom(0.1f, 0.2f, 0.2f, 0.3f, level);
				xDownNew = UnityEngine.Random.Range(0.7f, 1f - stroke);
				xUpNew = xDownNew + stroke;
				xWNew = ScaledRandom(1f, 2f, 2f, 4f, level);
				grindNew = ScaledRandom(-0.1f, 0.1f, -0.2f, 0.2f, level);
				tiltNew = ScaledRandom(0.2f, 0.4f, 0.3f, 0.6f, level);
				tiltWNew = ScaledRandom(-2.5f, 2.5f, -5f, 5f, level);
				vibe1New = ScaledRandom(0.1f, 1f, 0.3f, 1f, level);
			} else if (setMode == 4) {
				// Low Tease
				strokerReport.val = "Low Tease";
				stroke = ScaledRandom(0.1f, 0.2f, 0.2f, 0.3f, level);
				xDownNew = UnityEngine.Random.Range(0f, 0.1f);
				xUpNew = xDownNew + stroke;
				xWNew = ScaledRandom(0.5f, 1f, 1.5f, 3f, level);
				grindNew = ScaledRandom(-0.1f, 0.1f, -0.1f, 0.1f, level);
				tiltNew = ScaledRandom(0.5f, 0.8f, 0.5f, 0.9f, level);
				tiltWNew = ScaledRandom(-2f, 2f, -6f, 6f, level);
				vibe1New = ScaledRandom(-0.2f, 0.2f, -0.1f, 0.4f, level);
			} else if (setMode == 3) {
				// Top Tease
				strokerReport.val = "High Tease";
				stroke = ScaledRandom(0.1f, 0.2f, 0.2f, 0.3f, level);
				xDownNew = UnityEngine.Random.Range(0.7f, 1f - stroke);
				xUpNew = xDownNew + stroke;
				xWNew = ScaledRandom(0.5f, 1f, 1.5f, 3f, level);
				grindNew = ScaledRandom(-0.1f, 0.1f, -0.1f, 0.1f, level);
				tiltNew = ScaledRandom(0.5f, 0.8f, 0.5f, 0.9f, level);
				tiltWNew = ScaledRandom(-2f, 2f, -6f, 6f, level);
				vibe1New = ScaledRandom(-0.2f, 0.2f, -0.1f, 0.4f, level);
			} else if (setMode == 2) {
				// Short Strokes
				strokerReport.val = "Short Strokes";
				stroke = ScaledRandom(0.2f, 0.6f, 0.2f, 0.6f, level);
				xDownNew = UnityEngine.Random.Range(0f, 1f - stroke);
				xUpNew = xDownNew + stroke;
				xWNew = ScaledRandom(1.5f, 3f, 10f, 18f, level);
				grindNew = ScaledRandom(-0.3f, 0.3f, 0f, 0f, level);
				tiltNew = ScaledRandom(0f, 0.1f, 0f, 0.2f, level);
				tiltWNew = ScaledRandom(-0.5f, 0.5f, -1f, 1f, level);
				vibe1New = ScaledRandom(0f, 0f, 0f, 0f, level);
			} else {
				// Long Strokes
				strokerReport.val = "Long Strokes";
				stroke = ScaledRandom(0.6f, 1f, 0.8f, 1f, level);
				xDownNew = UnityEngine.Random.Range(0f, 1f - stroke);
				xUpNew = xDownNew + stroke;
				xWNew = ScaledRandom(0.5f, 2f, 3f, 5f, level);
				grindNew = ScaledRandom(-0.9f, 0.9f, -0.8f, 0.8f, level);
				tiltNew = ScaledRandom(0f, 0.1f, 0f, 0.2f, level);
				tiltWNew = ScaledRandom(-0.5f, 0.5f, -1f, 1f, level);
				vibe1New = ScaledRandom(0f, 0f, 0f, 0f, level);
			}
				
			// Calculate ramp rates
			xDUp = (xUpNew-xUp)/timer;
			xDDown = (xDownNew-xDown)/timer;
			xDW = (xWNew-xW)/timer;
			grindV = (grindNew - grind) / timer;
			tiltV = (tiltNew-tilt)/timer;
			tiltDW = (tiltWNew-tiltW)/timer;
			vibe1V = (vibe1New-vibe1)/timer;
		}
		
		// This function produces an appropriate random value to the intensity level
		float ScaledRandom(float lowMin, float lowMax, float highMin, float highMax, float level) {
			if (level < 0) level = 0;
			if (level > 1) level = 1;

			// Random function with interpolated inputs
			float output = UnityEngine.Random.Range(lowMin + (highMin-lowMin)*level,lowMax + (highMax-lowMax)*level);
			return output;
		}
		
		// Function to close plugin
        void OnDestroy() {
            StopSerial();
        }

		
    }
}